import pytest
import abow


@pytest.fixture(scope='module')
def monkeymodule():
    '''monkey patch with scope module
    '''
    from _pytest.monkeypatch import MonkeyPatch
    mpatch = MonkeyPatch()
    yield mpatch
    mpatch.undo()


def setup_backend_path(tmp_path_factory, monkeymodule):
    '''Create a temporary folder to hold the backend,
    and points to it via monkey patching config.

    Can also points to a non-accessible folder for testing error paths
    '''
    path = tmp_path_factory.mktemp('backend')
    cfg = abow.config.get

    def mock_cfg(key, default=None):
        '''override the config to set the temporary backend
        '''
        if key == 'abow.backend_path':
            return str(path)
        else:
            return cfg(key, None)

    monkeymodule.setattr(abow.config, "get", mock_cfg)

    return path


@pytest.fixture(scope="module")
def backend_path(tmp_path_factory, monkeymodule):
    '''usable backend
    '''
    setup_backend_path(tmp_path_factory, monkeymodule)


@pytest.fixture(scope="module")
def invalid_backend_path(tmp_path_factory, monkeymodule):
    '''non-accessible backend
    '''
    path = setup_backend_path(tmp_path_factory, monkeymodule)
    path.chmod(0)
    yield
    path.chmod(0o755)



import locale
import abow


def mock_cfg_valid(key):
    '''return a valid locale value
    '''
    return 'fr_FR.UTF-8'


def mock_cfg_invalid(key):
    '''return an invalid locale value
    '''
    return 'abcdef'


def test_deferred_gettext():
    '''test deferred gettext
    '''
    assert abow.l10n.deferred_gettext('abc') == 'abc'


def test_init(monkeypatch):
    '''test l10n initialisation and get_lang
    '''
    monkeypatch.setattr(abow.config, 'get', mock_cfg_valid)
    abow.l10n.init()
    assert abow.l10n.get_lang() == 'fr-FR'

    monkeypatch.setattr(abow.config, 'get', mock_cfg_invalid)
    abow.alerts.clear()
    abow.l10n.init()
    assert len(abow.alerts.get_list()) == 1


def test_get_lang():
    '''test get_lang with C locale
    '''
    cur_locale = locale.setlocale(locale.LC_ALL, 'C')
    assert abow.l10n.get_lang() == 'en'
    locale.setlocale(locale.LC_ALL, cur_locale)



import bottle
import abow


def test_print_file(capsys):
    '''test default configuration generation
    '''
    abow.config.print_file()
    cfg_file = capsys.readouterr().out
    assert '[abow]' in cfg_file


def test_keys():
    '''make sure the expected config keys are here
    '''
    cfg = bottle.ConfigDict()
    abow.config.init(cfg)
    keys = [
        'abow.locale_path',
        'abow.static_path',
        'abow.template_path',
        'abow.backend_path',
        'abow.static_url',
        'abow.locale',
        'abow.welcome_page',
        'abow.help_page',
        ]
    assert len(keys) == len(cfg.keys())
    assert sorted(keys) == sorted(cfg.keys())


def test_load_cfg(monkeypatch, tmp_path):
    '''test configuration load from file
    '''
    # from XDG_CONFIG_HOME
    user_conf_dir = tmp_path / 'abow'
    user_conf_dir.mkdir()
    user_conf_file = user_conf_dir / 'config'
    user_conf_file.write_text('[abow]\nwelcome_page=abc\nhelp_page=def\n')
    monkeypatch.setenv('XDG_CONFIG_HOME', str(tmp_path))

    cfg = bottle.ConfigDict()
    abow.config.init(cfg)
    assert abow.config.get('abow.welcome_page') == 'abc'
    assert abow.config.get('abow.help_page') == 'def'

    # from ABOW_CONFIG
    user_conf_file = tmp_path / 'abow.ini'
    user_conf_file.write_text('[abow]\nwelcome_page=xyz\n')
    monkeypatch.setenv('ABOW_CONFIG', str(user_conf_file))

    cfg = bottle.ConfigDict()
    abow.config.init(cfg)
    assert abow.config.get('abow.welcome_page') == 'xyz'
    assert abow.config.get('abow.help_page') == 'def'



import abow


def test_init():
    '''No alert at startup
    '''
    assert abow.alerts.get_list() == []


def test_add():
    '''add some alerts
    '''
    abow.alerts.add('info', '1')
    abow.alerts.add('danger', '2')
    assert abow.alerts.get_list() == [('info', '1'), ('danger', '2')]


def test_clear():
    '''clear alerts
    '''
    abow.alerts.clear()
    assert abow.alerts.get_list() == []



import textwrap
import abow


def test_empty():
    '''test empty string
    '''
    assert abow.render.render('') == {'html': '', 'toc': ''}


def test_text():
    '''make sure text is rendered
    Note: we are not testing markdown here
    '''
    text = 'abc'
    result = abow.render.render(text)
    assert text in result['html']
    assert result['toc'] != ''


def test_wikilinks():
    '''test the wikilinks extension
    '''
    # normal wiki link
    result = abow.render.render('[[label|target]]')
    assert result['html'] == \
        '<p><a class="wikilink text-danger" href="../view/target">label</a></p>'

    # action link
    result = abow.render.render('[[label|action:edit/target]]')
    assert result['html'] == '<p><a class="wikilink" href="../edit/target">label</a></p>'

    # invalid action link
    result = abow.render.render('[[action:]]')
    assert result['html'] == '<p></p>'

    # anchor link
    result = abow.render.render('[[#abc]]')
    assert result['html'] == '<p><a class="wikilink" href="#abc">#abc</a></p>'


def test_table():
    '''test the tables extension
    '''
    result = abow.render.render('| a | b |\n|---|---|\n| c | d |')
    expected = '''\
        <div class="table-responsive">
        <table class="table table-bordered">
        <thead class="thead-light">
        <tr>
        <th>a</th>
        <th>b</th>
        </tr>
        </thead>
        <tbody>
        <tr>
        <td>c</td>
        <td>d</td>
        </tr>
        </tbody>
        </table>
        </div>'''
    assert result['html'] == textwrap.dedent(expected)



import pytest
import abow


pytestmark = pytest.mark.usefixtures('backend_path')


@pytest.fixture(scope="module")
def create_pages():
    '''pages must exist to search for tags
    '''
    abow.backend.set_content('page1', 'abc')
    abow.backend.set_content('page2', 'abc')


def test_set_tag():
    '''test setting and getting tags
    '''
    assert abow.tags.get_tags('page1') == []
    abow.tags.set_tags('page1', [])
    assert abow.tags.get_tags('page1') == []

    abow.tags.set_tags('page1', ['b', 'b', 'a'])
    assert abow.tags.get_tags('page1') == ['b', 'a']

    abow.tags.set_tags('page1', 'abc, def; gh, i j k , ; ,zzz  ')
    assert abow.tags.get_tags('page1') == ['abc', 'def', 'gh', 'i j k', 'zzz']

    with pytest.raises(TypeError):
        abow.tags.set_tags('page1', 0)


def test_search(create_pages):
    '''test searching by tag
    '''
    abow.tags.set_tags('page1', ['a', 'b'])
    abow.tags.set_tags('page2', ['a', 'c'])

    assert abow.tags.search('a') == ['page1', 'page2']
    assert abow.tags.search('b') == ['page1']
    assert abow.tags.search('c') == ['page2']
    assert abow.tags.search('d') == []


def test_get_all_tags(create_pages):
    '''test retreiving all the tags
    '''
    abow.tags.set_tags('page1', ['a', 'b'])
    abow.tags.set_tags('page2', ['a', 'c'])

    assert abow.tags.get_all_tags() == ['a', 'b', 'c']



# A Bottle of Wiki — personal wiki
# SPDX-License-Identifier: AGPL-3.0-or-later
# Copyright (C) 2020  Benoît Monin <benoit.monin@gmx.fr>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from collections import OrderedDict
import locale
import re

from . import backend


# attribute name for the tags
_ATTR = 'tags'


def get_tags(pagename):
    '''Return the list of tags of a page
    '''
    return backend.get_attr(pagename, _ATTR, [])


def set_tags(pagename, tags):
    '''Set the tags of a page
    '''
    if isinstance(tags, str):
        # convert to list by spliting on ; and ,
        tags = [t.strip() for t in re.split('[;,]', tags)]
    if not isinstance(tags, list):
        raise TypeError('tags must be a string or a list')

    # remove empty tags and duplicates but keep order
    tags = list(OrderedDict.fromkeys(filter(None, tags)))

    if get_tags(pagename) or tags:
        backend.set_attr(pagename, _ATTR, tags)


def search(tag):
    '''Search pages matching a tag and return a list

    Note: the list is sorted
    '''
    tagged_pages = []
    all_pages = backend.get_pages_list()

    for page in all_pages:
        tags = get_tags(page)
        if tag in tags:
            tagged_pages.append(page)

    return tagged_pages


def get_all_tags():
    '''Return the list of the tags of all pages

    Note: the list is sorted
    '''
    tags = set()
    pages = backend.get_pages_list()

    for page in pages:
        tags.update(get_tags(page))

    return sorted(tags, key=locale.strxfrm)



# A Bottle of Wiki — personal wiki
# SPDX-License-Identifier: AGPL-3.0-or-later
# Copyright (C) 2020  Benoît Monin <benoit.monin@gmx.fr>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

try:
    from importlib import resources
except ImportError:
    import importlib_resources as resources
import os
import os.path
import locale
import textwrap

# keep a reference of the configuration
_cfg = None


def _resource_path(res):
    '''Return the absolute path of a package resource

    Note: this only works if the package is not compressed
    '''
    with resources.path(__package__, res) as p:
        return str(p)


def _set_default(cfg):
    '''Set the default configuration values
    '''

    # not shown to the user, internal paths
    cfg.setdefault('abow.locale_path', _resource_path('locale'))
    cfg.setdefault('abow.static_path', _resource_path('static'))
    cfg.setdefault('abow.template_path', _resource_path('views'))

    cfg.meta_set('abow.backend_path', 'help', __('Path to the backend storage'))
    cfg.meta_set('abow.backend_path', 'extra_help',
                 'The backend is where A Bottle of Wiki stores the pages content '
                 'and attributes. It should points to a directory writable by '
                 'the user running the application.')
    cfg.meta_set('abow.backend_path', 'dflt_help', 'the current directory')
    cfg.setdefault('abow.backend_path', os.path.abspath('.'))

    cfg.meta_set('abow.static_url', 'help', __('Base url for the static resources, '
                 'if hosted outside of the app'))
    cfg.meta_set('abow.static_url', 'extra_help',
                 'Define this url if you want to serve the static resources '
                 'directly from a webserver instead of through the application.')
    cfg.meta_set('abow.static_url', 'dflt_help', 'blank, static resources are hosted by the app')
    cfg.setdefault('abow.static_url', '')

    cfg.meta_set('abow.locale', 'help', __('The locale to use'))
    cfg.meta_set('abow.locale', 'extra_help',
                 'Set this value to override the locale settings. '
                 'English and french localization are available.')
    cfg.meta_set('abow.locale', 'dflt_help', 'the current locale from the environnment variables')
    cfg.setdefault('abow.locale', locale.getdefaultlocale()[0] or 'en_US')

    cfg.meta_set('abow.welcome_page', 'help', __('Name of the welcome page '
                 'of the wiki'))
    welcome_page = 'welcome'
    cfg.meta_set('abow.welcome_page', 'extra_help',
                 'The welcome page is the page you reach when connecting '
                 'to the root of the application.')
    cfg.meta_set('abow.welcome_page', 'dflt_help', welcome_page)
    cfg.setdefault('abow.welcome_page', welcome_page)

    cfg.meta_set('abow.help_page', 'help', __('Name of the help page '
                 'of the wiki'))
    help_page = 'usage'
    cfg.meta_set('abow.help_page', 'extra_help',
                 'The help page, if set, is available as a menu entry in the '
                 'navigation bar.')
    cfg.meta_set('abow.help_page', 'dflt_help', help_page)
    cfg.setdefault('abow.help_page', help_page)


def _load_cfg_files(cfg):
    '''Load the configuration files one by one
    '''
    # global configuration
    cfg.load_config(os.path.join('/', 'etc', 'abow', 'config'))

    # user configuration
    if 'XDG_CONFIG_HOME' in os.environ:
        path = [os.path.expandvars('$XDG_CONFIG_HOME')]
    else:
        path = [
            os.path.expanduser('~'),
            '.config'
            ]
    cfg.load_config(os.path.join(*path, 'abow', 'config'))

    # specific configuration
    if 'ABOW_CONFIG' in os.environ:
        cfg.load_config(os.path.expandvars('$ABOW_CONFIG'))


def init(cfg):
    '''Initialize the configuration dictionary

    Set the default value
    '''
    global _cfg
    _cfg = cfg

    _set_default(cfg)
    _load_cfg_files(cfg)


def get(key, default=None):
    '''Return the value of the configuration or the default value
    '''
    global _cfg
    return _cfg.get(key, default)


def print_file():
    '''Print the default configuration file to stdout

    Note: this function is meant to be used independently of the app
    '''
    from bottle import ConfigDict
    cfg = ConfigDict()
    _set_default(cfg)

    # wrapper for multiline comments
    wrapper = textwrap.TextWrapper(initial_indent='# ', subsequent_indent='# ')

    # header
    header = '''\
        -- A Bottle of Wiki configuration file --

        The configuration is read from the following paths:
         * /etc/abow/config
         * $XDG_CONFIG_HOME/abow/config, defaulting to ~/.config/abow/config
         * $ABOW_CONFIG, if defined
        Paths are tried one after one and each configuration file can
        override the settings of the previous ones.
        The settings used can be viewed in the 'about' page.
        '''
    print(textwrap.indent(textwrap.dedent(header), '# ', lambda line: True))
    print('[abow]')

    # the user-modifiable configuration options (i.e. with help text)
    for k in sorted(cfg.keys()):
        if k.startswith('abow.') and cfg.meta_get(k, 'help'):
            print('\n##', cfg.meta_get(k, 'help'))
            print(wrapper.fill(cfg.meta_get(k, 'extra_help', '-')))
            print('# Default value if unset:', cfg.meta_get(k, 'dflt_help'))
            print('#{} ='.format(k.split('.')[1]))



import io
import bottle
import pytest
import abow
try:
    from importlib import metadata
except ImportError:
    import importlib_metadata as metadata


pytestmark = pytest.mark.usefixtures('backend_path')


def test_init():
    '''test app initialisation
    '''
    abow.alerts.clear()
    bottle.DEBUG = True
    abow.app.init()
    assert len(abow.alerts.get_list()) == 1
    bottle.DEBUG = False


def test_is_modified(monkeypatch):
    '''test is modified return value
    '''
    # empty backend
    assert abow.app._is_modified() is True

    # debug mode
    bottle.DEBUG = True
    assert abow.app._is_modified() is True
    bottle.DEBUG = False

    # create a page to have an backend mtime
    abow.backend.set_content('page1', 'abc')
    mtime = abow.backend.get_mtime()

    assert abow.app._is_modified() is True
    assert 'Last-Modified' in bottle.response.headers

    # simulate sending if modified since
    monkeypatch.setitem(bottle.request.environ,
                        'HTTP_IF_MODIFIED_SINCE',
                        bottle.http_date(mtime))
    assert abow.app._is_modified() is False

    monkeypatch.setitem(bottle.request.environ,
                        'HTTP_IF_MODIFIED_SINCE',
                        bottle.http_date(mtime - 1))
    assert abow.app._is_modified() is True


def test_get_url(monkeypatch):
    '''test url generation
    '''
    assert abow.app.get_url('root') == '/'
    assert abow.app.get_url('view', pagename='page1') == '/view/page1'

    def mock_cfg(key):
        return None
    monkeypatch.setattr(abow.config, 'get', mock_cfg)
    assert abow.app.get_url('static', filepath='test') == '/static/test'

    def mock_cfg(key):
        return '/outside'
    monkeypatch.setattr(abow.config, 'get', mock_cfg)
    assert abow.app.get_url('static', filepath='test') == '/outside/test'


def test_init_tpl_response():
    '''test the presence of the common keys
    '''
    expected = ['action', 'alerts', 'get_url', 'help_page', 'lang', 'title']
    resp = abow.app.init_tpl_response('abc', 'def')

    assert resp['action'] == 'abc'
    assert resp['title'] == 'def'
    assert len(expected) == len(resp.keys())
    assert sorted(expected) == sorted(resp.keys())


def test_static():
    '''test access to static resources
    '''
    assert abow.app.get_static(filepath='not-found').status_code == 404
    assert abow.app.get_static(filepath='css/abow.css').status_code == 200


def test_redirect_to_index(monkeypatch):
    '''test redirection to the wiki home page
    '''
    def mock_cfg(key):
        return 'welcome'
    monkeypatch.setattr(abow.config, 'get', mock_cfg)

    with pytest.raises(bottle.HTTPResponse) as resp:
        abow.app.redirect_to_index()

    assert resp.value.status_code == 302
    assert 'view/welcome' in resp.value.headers['Location']


def test_view_page():
    '''test the view page
    '''
    pagename = 'abc'
    data = abow.app.view_page.__wrapped__(pagename)
    assert data['pagename'] == pagename
    assert data['title'] == pagename
    assert data['action'] == 'view'
    assert {'tags', 'html', 'toc'}.issubset(set(data.keys()))


def test_edit_page():
    '''test the edit page
    '''
    pagename = 'abc'
    data = abow.app.edit_page.__wrapped__(pagename)
    assert data['pagename'] == pagename
    assert pagename in data['title']
    assert data['action'] == 'edit'
    assert {'tags', 'all_tags', 'content'}.issubset(set(data.keys()))


def test_list_pages(monkeypatch):
    '''test the list pages
    '''
    data = abow.app.list_pages.__wrapped__()
    assert data['action'] == 'list_pages'
    assert 'pages_list' in data.keys()

    monkeypatch.setitem(bottle.request.environ,
                        'bottle.request.query',
                        bottle.FormsDict({'text': 'abc'}))
    data = abow.app.list_pages.__wrapped__()
    assert data['action'] == 'list_pages'
    assert data['pagename'] == 'abc'
    assert {'pages_list', 'page_exists', 'pagename'}.issubset(data.keys())

    monkeypatch.setitem(bottle.request.environ,
                        'bottle.request.query',
                        bottle.FormsDict({'text': 'abc/def'}))
    data = abow.app.list_pages.__wrapped__()
    assert 'pages_list' in data.keys()
    assert {'page_exists', 'pagename'}.isdisjoint(data.keys())

    data = abow.app.list_pages.__wrapped__(tag='a')
    assert data['action'] == 'list_pages'
    assert 'pages_list' in data.keys()


def test_about(monkeypatch):
    '''test the about page
    '''
    data = abow.app.about.__wrapped__()
    assert data['action'] == 'about'
    assert {'version', 'config'}.issubset(set(data.keys()))

    def mock_get(key):
        raise metadata.PackageNotFoundError
    monkeypatch.setattr(metadata, 'version', mock_get)
    data = abow.app.about.__wrapped__()


def test_not_modified(monkeypatch):
    '''test the 304 reply
    '''
    # create a page to have an backend mtime
    abow.backend.set_content('page1', 'abc')
    mtime = abow.backend.get_mtime()

    # simulate sending if modified since
    monkeypatch.setitem(bottle.request.environ,
                        'HTTP_IF_MODIFIED_SINCE',
                        bottle.http_date(mtime))

    routes = [
        abow.app.view_page,
        abow.app.edit_page,
        abow.app.list_pages,
        abow.app.about,
        ]

    for route in routes:
        assert route.__wrapped__('').status_code == 304


def test_error():
    '''test the error page
    '''
    err = bottle.HTTPError(404)
    data = abow.app.error.__wrapped__(err)
    assert data['action'] == 'error'
    assert 'err_body' in data.keys()


def test_save_page(monkeypatch):
    '''test the save page
    '''
    form = bottle.FormsDict()
    form['content'] = 'abc'
    form['new_pagename'] = 'page1'
    form['tags'] = 'tag1'
    monkeypatch.setitem(bottle.request.environ,
                        'bottle.request.forms',
                        form)

    with pytest.raises(bottle.HTTPResponse) as resp:
        abow.app.save_page('page1')
    assert resp.value.status_code == 302
    assert 'view/page1' in resp.value.headers['Location']

    with pytest.raises(bottle.HTTPResponse):
        abow.app.save_page('page2')

    form['new_pagename'] = 'page1/1'
    with pytest.raises(bottle.HTTPError) as err:
        abow.app.save_page('page1')
    assert err.value.status_code == 403

    form['new_pagename'] = 'page1|1'
    with pytest.raises(bottle.HTTPError) as err:
        abow.app.save_page('page1')
    assert err.value.status_code == 403

    form['new_pagename'] = 'page1'
    with pytest.raises(bottle.HTTPError) as err:
        abow.app.save_page('page2')
    assert err.value.status_code == 403

    form['new_pagename'] = 'page3'
    with pytest.raises(bottle.HTTPResponse) as resp:
        abow.app.save_page('page1')
    assert resp.value.status_code == 302
    assert 'view/page3' in resp.value.headers['Location']


def test_preview(monkeypatch):
    '''test the preview page
    '''
    monkeypatch.setitem(bottle.request.environ,
                        'CONTENT_TYPE',
                        'text/plain')
    monkeypatch.setitem(bottle.request.environ,
                        'bottle.request.body',
                        io.BytesIO(b'abc'))
    res = abow.app.preview()
    assert res['html'] == '<p>abc</p>'

    monkeypatch.setitem(bottle.request.environ,
                        'CONTENT_TYPE',
                        'text/plain; charset=UTF-8')
    res = abow.app.preview()
    assert res['html'] == '<p>abc</p>'



# A Bottle of Wiki — personal wiki
# SPDX-License-Identifier: AGPL-3.0-or-later
# Copyright (C) 2020  Benoît Monin <benoit.monin@gmx.fr>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import re
import urllib.parse
import builtins
import math
import bottle
try:
    from importlib import metadata
except ImportError:
    import importlib_metadata as metadata

from . import alerts
from . import backend
from . import config
from . import l10n
from . import render
from . import tags

# instanciate the bottle application
app = bottle.Bottle()


def init():
    '''Initialisation of the application
    '''
    # globally setup the deferred gettext
    builtins.__dict__['__'] = l10n.deferred_gettext

    config.init(app.config)
    l10n.init()
    bottle.TEMPLATE_PATH.insert(0, config.get('abow.template_path'))

    if bottle.DEBUG:
        alerts.add('info', __('Application started in debug mode.'))


# setup the app
init()


def _is_modified():
    '''Manage the Last-Modified and If-Modified-Since headers

    Avoid recomputing and resending a page if already cached by the browser
    '''
    # make sure the browser don't use stall data
    bottle.response.set_header('Cache-Control', 'no-cache')

    # no caching when debugging
    if bottle.DEBUG:
        return True

    last_modified = backend.get_mtime()
    ims = bottle.request.environ.get('HTTP_IF_MODIFIED_SINCE')
    if ims:
        ims = bottle.parse_date(ims.split(";")[0].strip())
    if ims is not None and ims >= math.floor(last_modified):
        return False

    last_modified = bottle.http_date(last_modified)
    bottle.response.set_header('Last-Modified', last_modified)

    return True


def get_url(routename, **kwargs):
    '''Local version of get_url to manage static resources hosted outside
    of the app
    '''
    if routename == 'static':
        static_url = config.get('abow.static_url')
        if static_url:
            return urllib.parse.urljoin(static_url + '/', kwargs['filepath'])
    return app.get_url(routename, **kwargs)


def init_tpl_response(action='', title=''):
    '''Create a dictionary with the minimal info needed to render a template
    '''
    return {
        'get_url': app.get_url,
        'alerts': alerts,
        'help_page': config.get('abow.help_page'),
        'lang': l10n.get_lang(),
        'action': action,
        'title': title,
        }


@app.get('/static/<filepath:path>', name='static')
def get_static(filepath):
    '''Serve the static resources
    '''
    return bottle.static_file(filepath, root=config.get('abow.static_path'))


@app.get('/', name='root')
def redirect_to_index():
    '''Redirect to the index page, the main page of the wiki
    '''
    welcome_page = config.get('abow.welcome_page')
    return bottle.redirect('view/' + welcome_page, code=302)


@app.get('/view/<pagename>', name='view')
@bottle.view('view')
def view_page(pagename):
    '''Display a page of the wiki
    '''
    if not _is_modified():
        return bottle.HTTPResponse(status=304)

    data = init_tpl_response('view', pagename)
    data['pagename'] = pagename
    data['tags'] = tags.get_tags(pagename)
    data.update(render.render(backend.get_content(pagename)))

    return data


@app.get('/edit/<pagename>', name='edit')
@bottle.view('edit')
def edit_page(pagename):
    '''Edit a page of the wiki
    '''
    if not _is_modified():
        return bottle.HTTPResponse(status=304)

    data = init_tpl_response('edit', _('Edit {}').format(pagename))
    data.update({
        'pagename': pagename,
        'tags': '; '.join(tags.get_tags(pagename)),
        'all_tags': tags.get_all_tags(),
        'content': backend.get_content(pagename),
        })
    return data


@app.get('/tag/<tag>')
@app.get('/search/pages', name='search')
@bottle.view('list_pages')
def list_pages(tag=None):
    '''Display the list of pages in the wiki
    optionally filtered by tag or text search
    '''
    if not _is_modified():
        return bottle.HTTPResponse(status=304)

    data = init_tpl_response('list_pages')
    text = bottle.request.query.getunicode('text')
    if tag:
        data['title'] = _('List of pages tagged {}').format(tag)
        data['pages_list'] = tags.search(tag)
    elif text:
        data['title'] = _('List of pages containing {}').format(text)
        data['pages_list'] = backend.search_pages(text)
        if '/' not in text:
            data['page_exists'] = text in backend.get_pages_list()
            data['pagename'] = text
    else:
        data['title'] = _('List of pages')
        data['pages_list'] = backend.get_pages_list()

    return data


@app.get('/about', name='about')
@bottle.view('about')
def about(tag=None):
    '''Display the about page
    '''
    if not _is_modified():
        return bottle.HTTPResponse(status=304)

    data = init_tpl_response('about', _('About'))
    try:
        data['version'] = _('Version {}').format(metadata.version(__package__))
    except metadata.PackageNotFoundError:
        data['version'] = _('Developement version')
    data['config'] = app.config
    return data


@app.error(403)
@app.error(404)
@app.error(405)
@app.error(500)
@bottle.view('error')
def error(err):
    '''Display the various http errors with the wiki style
    '''
    data = init_tpl_response('error', _('Error {}').format(err.status_code))
    data['err_body'] = err.body
    return data


@app.post('/edit/<pagename>')
def save_page(pagename):
    '''Save the page from the edit form
    '''
    content = bottle.request.forms.getunicode('content')
    new_pagename = bottle.request.forms.getunicode('new_pagename')
    page_tags = bottle.request.forms.getunicode('tags')

    if not new_pagename or re.search(r'[/|]', new_pagename):
        bottle.abort(403, _('Page name cannot be blank and cannot contain \'/\' or \'|\''))

    if new_pagename != pagename and new_pagename in backend.get_pages_list():
        bottle.abort(403, _('A page named {} already exists').format(new_pagename))

    tags.set_tags(new_pagename, page_tags)
    backend.set_content(new_pagename, content)
    if new_pagename != pagename:
        backend.delete(pagename)

    return bottle.redirect('../view/' + urllib.parse.quote(new_pagename))


@app.put('/preview')
def preview():
    '''Render the passed-in text and return the html fragment generated
    '''
    content_type = bottle.request.content_type
    if 'charset=' in content_type:
        encoding = content_type.split('charset=')[-1].split(';')[0].strip()
    else:
        encoding = 'utf8'

    content = bottle.request.body.read()
    content = content.decode(encoding)
    return render.render(content)
